//=============================================================================//
//	Joystick Input Module
//	Version 1.0
//	Written by Night-Eagle
//=============================================================================//

#include <windows.h>
#include "common/GMLuaModule.h"
#include <dinput.h>
#include <iostream>
#include <stdio.h>

//#include <atlbase.h>
//#include <comutil.h> //_bsrt_t

//#define DIRECTINPUT_VERSION 0x0700

// Libraries
//#pragma comment (lib, "dinput.lib")
#pragma comment (lib, "dinput8.lib")
#pragma comment (lib, "dxguid.lib")
#pragma comment (lib, "ole32.lib")

//using namespace std;
//using namespace System;

// Globals
LPDIRECTINPUT8 din; //Root DirectInput Interface
LPDIRECTINPUTDEVICE8 dinkeyboard; //Keyboard Device
const int max_devices = 64; //Maximum number of devices

// Joystick globals
LPDIRECTINPUTDEVICE8 joy_din[max_devices]; //Joystick Device
DIDEVCAPS joy_devcaps[max_devices]; //DIDEVCAPS Struct
//GUID joy_guid[max_devices]; //GUID for joystick
char joy_name[max_devices][MAX_PATH]; //Name of joystick
float joy_guid[max_devices][3]; //GUID of joystick 1
float joy_guids[max_devices][8]; //GUID of joystick 2
//char joy_guid[max_devices][MAX_PATH]; //GUID of joystick
int joy_n = 0;
char* binaryversion = "1.1";

DIJOYSTATE2 joy_state[max_devices]; //State of joystick
bool joy_active[max_devices]; //Joystick present
// End joystick globals

GMOD_MODULE( Init, Shutdown );

//
// Lua Functions
//

LUA_FUNCTION( refresh )
{
	int joy = (g_Lua -> GetInteger(1));

	if ((joy > 0)&&(joy <=joy_n))
	{
		HRESULT hr;
		//Msg("Active?\n");
		if (joy_din[joy]) //(joy_active[joy])
		{
			//Msg("Yes...");
			if (FAILED(hr = joy_din[joy]->Poll()))
			{
				//Msg("Poll fail\n");
				hr = joy_din[joy]->Acquire();
			}
			else
			{
				//Msg("Poll good\n");
				hr = joy_din[joy]->GetDeviceState(
					sizeof(DIJOYSTATE2),
					&joy_state[joy]
				);
			}
		}
	}
	else
	{
		int i;
		for (i=0;i<joy_n;i++)
		{
			HRESULT hr;
			//Msg("Active?\n");
			if (joy_din[i]) //(joy_active[i])
			{
				//Msg("Yes...");
				if (FAILED(hr = joy_din[i]->Poll()))
				{
					//Msg("Poll fail\n");
					hr = joy_din[i]->Acquire();
				}
				else
				{
					//Msg("Poll good\n");
					hr = joy_din[i]->GetDeviceState(
						sizeof(DIJOYSTATE2),
						&joy_state[i]
					);
				}
			}
		}
	}
	return 0;
}

LUA_FUNCTION( axis )
{
	int joy = (g_Lua -> GetInteger(1));
	int axi = (g_Lua -> GetInteger(2));
	long out;

	switch(axi)
	{
	case 0 :
		out = joy_state[joy].lX;
		break;
	case 1 :
		out = joy_state[joy].lY;
		break;
	case 2 :
		out = joy_state[joy].lZ;
		break;
	case 3 :
		out = joy_state[joy].lRx;
		break;
	case 4 :
		out = joy_state[joy].lRy;
		break;
	case 5 :
		out = joy_state[joy].lRz;
		break;
	case 6 :
		out = joy_state[joy].rglSlider[0];
		break;
	case 7 :
		out = joy_state[joy].rglSlider[1];
		break;
	default :
		out = 0;
	}
	
	g_Lua->Push( (float)out );
	return 1;
}

LUA_FUNCTION( button )
{
	int joy = (g_Lua -> GetInteger(1));
	int but = (g_Lua -> GetInteger(2));
	g_Lua->Push( (float)(joy_state[joy].rgbButtons[but]) );
	return 1;
}

LUA_FUNCTION( pov )
{
	int joy = (g_Lua -> GetInteger(1));
	int pov = (g_Lua -> GetInteger(2));
	g_Lua->Push( (float)(joy_state[joy].rgdwPOV[pov]) );
	return 1;
}

LUA_FUNCTION( count )
{
	int joy = (g_Lua -> GetInteger(1));
	int opt = (g_Lua -> GetInteger(2));
	float out;
	switch(opt)
	{
	case 1 :
		out = (float)joy_devcaps[joy].dwAxes;
		break;
	case 2 :
		out = (float)joy_devcaps[joy].dwButtons;
		break;
	case 3 :
		out = (float)joy_devcaps[joy].dwPOVs;
		break;
	default :
		out = (float)joy_n;
	}
	g_Lua->Push(out);
	return 1;
}

LUA_FUNCTION( name )
{
	int joy = (g_Lua -> GetInteger(1));
	g_Lua->Push( (char*) joy_name[joy] );
	return 1;
}

LUA_FUNCTION( guidm )
{
	int joy = (g_Lua -> GetInteger(1));
	g_Lua->Push( (float) joy_guid[joy][0] );
	g_Lua->Push( (float) joy_guid[joy][1] );
	g_Lua->Push( (float) joy_guid[joy][2] );
	int i;
	for (i=0;i<8;i++){
		g_Lua->Push( (float) joy_guids[joy][i] );
	}
	return 11;
}


//
// Axis Initialization
//

BOOL CALLBACK EnumAxesCallback(const DIDEVICEOBJECTINSTANCE* pdidoi, VOID* pContext)
{
	
	HRESULT hr;
	DIPROPRANGE diprg; 
	
	
	diprg.diph.dwSize       = sizeof(DIPROPRANGE); 
	diprg.diph.dwHeaderSize = sizeof(DIPROPHEADER); 
	diprg.diph.dwHow        = DIPH_BYID; 
	diprg.diph.dwObj        = pdidoi->dwType;
	if ((pdidoi->dwType)&DIDFT_AXIS)
	{
		diprg.lMin              = 0; 
		diprg.lMax              = +65535;
	}
	else
	{
		diprg.lMin              = 0; 
		diprg.lMax              = +1;
	}
	
	if (FAILED(hr = joy_din[joy_n]->SetProperty(
		DIPROP_RANGE,
		&diprg.diph
	)))
	{
		return DIENUM_STOP;
	}
	else
	{
		return DIENUM_CONTINUE;
	}
}

//
// Joystick Initialization
//

BOOL CALLBACK EnumJoysticksCallback(const DIDEVICEINSTANCE* pdidInstance, VOID* pContext)
{
	HRESULT hr;

	// Get interface
	hr = din->CreateDevice(
		pdidInstance->guidInstance,
		//&g_pJoystick,
		&joy_din[joy_n],
		NULL
	);
	
	if (FAILED(hr))
	{
		//Msg("JOYENUM: Bad joystick.\n");
		return DIENUM_CONTINUE;
	}
	//Msg("JOYENUM: Found joystick...Setting Data Format...");
	if (FAILED(hr = joy_din[joy_n]->SetDataFormat(&c_dfDIJoystick2)))
	{
		//Msg("Set data format failed!\n");
		return DIENUM_CONTINUE;
	}
	//Msg("Setting cooperative level...");
	if (FAILED(hr = joy_din[joy_n]->SetCooperativeLevel(
		NULL,
		DISCL_NONEXCLUSIVE | DISCL_BACKGROUND
	)))
	{
		//Msg("Set cooperative level failed!\n");
		return DIENUM_CONTINUE;
	}
	//Msg("DIDEVCAPS...");
	joy_devcaps[joy_n].dwSize = sizeof(DIDEVCAPS);
	if (FAILED(joy_din[joy_n]->GetCapabilities(
		&joy_devcaps[joy_n]
	)))
	{
		//Msg("DIDEVCAPS failed!\n");
		return DIENUM_CONTINUE;
	}
	//Msg("Success!\nEnumerating axes...");
	if (FAILED(joy_din[joy_n]->EnumObjects(
		EnumAxesCallback,
		NULL,
		NULL
		//DIDFT_AXIS
	)))
	{
		//Msg("Failed!\n");
		return DIENUM_CONTINUE;
	}
	
	// User stuff
	if ((joy_devcaps[joy_n].dwFlags & DIDC_ATTACHED) > 0)
	{
		//Msg("Joystick is attached. ");
		joy_active[joy_n] = true;
	}

	//Get device info
	int i;
	for (i=0;i<MAX_PATH;i++)
	{
		joy_name[joy_n][i] = pdidInstance->tszProductName[i];
	}
	
	joy_guid[joy_n][0] = (pdidInstance->guidInstance).Data1;
	joy_guid[joy_n][1] = (pdidInstance->guidInstance).Data2;
	joy_guid[joy_n][2] = (pdidInstance->guidInstance).Data3;
	for (i=0;i<8;i++)
	{
		joy_guids[joy_n][i] = (pdidInstance->guidInstance).Data4[i];
	}
	
	// Acquire the joystick
	joy_din[joy_n]->Acquire();
	
	// Increment joy_n
	joy_n++;
	//Msg("Prepared for next joystick...\n\n");

	// We want to enumerate all joysticks, so keep on enumerating until we are out of joysticks
	return DIENUM_CONTINUE;
}


//
// Initialization
//

int Init( void )
{
	Msg("Joystick module loaded!\n1.0 Night-Eagle\n");
	//Msg("Night-Eagle's joystick module loading...\n");
	
	// Register our functions as globals
	g_Lua->SetGlobal( "refresh", refresh );
	g_Lua->SetGlobal( "axis", axis );
	g_Lua->SetGlobal( "button", button );
	g_Lua->SetGlobal( "pov", pov );
	g_Lua->SetGlobal( "count", count );
	g_Lua->SetGlobal( "name", name );
	g_Lua->SetGlobal( "guidm", guidm );
	g_Lua->SetGlobal( "binaryversion", binaryversion );

	// Lets make a global table filled with functions (a module)
	g_Lua->NewGlobalTable( "joystick" );
	ILuaObject* table = g_Lua->GetGlobal( "joystick" );

		table->SetMember( "refresh", refresh );
		table->SetMember( "axis", axis );
		table->SetMember( "button", button );
		table->SetMember( "pov", pov );
		table->SetMember( "count", count );
		table->SetMember( "name", name );
		table->SetMember( "guidm", guidm );
		table->SetMember( "binaryversion", binaryversion );

	// Delete table reference
	SAFE_UNREF( table );
	
	// Start DirectInput
	//Msg("Initializing DirectInput8...");
	DirectInput8Create(
		GetModuleHandle(NULL),
		DIRECTINPUT_VERSION,
		IID_IDirectInput8,
		(void**)&din,
		NULL
	);

	if (din == NULL)
	{
		//Msg("Failed!\n");
		return false;
	}
	else
	{
		//Msg("Success!\n");
	}
	
	//Msg("Fetching keyboard...");
	din->CreateDevice(
		GUID_SysKeyboard,
		&dinkeyboard,
		NULL
	);
	
	if (dinkeyboard)
	{
		dinkeyboard->SetDataFormat(&c_dfDIKeyboard);
		dinkeyboard->SetCooperativeLevel(
			//this->m_hWnd,
			NULL,
			DISCL_FOREGROUND | DISCL_EXCLUSIVE
		);
		dinkeyboard->Acquire();
		//Msg("Success!\n");
	}
	else
	{
		//Msg("Failed!\n");
		return false;
	}
	
	//Release the keyboard (All this was just for fun!!!)
	//Msg("Unacquiring keyboard.\n");
	dinkeyboard->Unacquire();
	
	//
	// Joystick Meat
	//
	
	//Msg("Enumerating joysticks...\n");
	din->EnumDevices(
		DI8DEVCLASS_GAMECTRL,
		EnumJoysticksCallback,
		NULL,
		DIEDFL_ATTACHEDONLY
	);
	//Msg("Joysticks enumerated!\n");
	
	return 0;
}

// When the DLL is shutdown it calls this function.
// g_Lua should be assumed to have been released here
int Shutdown( void )
{
	//Close DirectInput
	//Msg("Closing DirectInput\n");
	din->Release();
	return 0;
}

